"""Usage example"""
# To test Library installed from PyPi, use this import:

from ontu_parser.classes import Parser

# To test Library from source code, use this import:

# from classes import Parser

# An example of how to pass arguments to Firefox browser created by selenium
# Add arguments inside 'browser_settings' and see what happens
parser = Parser()

schedule = parser.parse(all_time=False)
for day_name, pairs in schedule.items():
    print(f"{day_name}:\n")
    for pair in pairs:
        if not pair.lessons:
            continue
        print(f"{pair.pair_no}:")
        for lesson in pair.lessons:
            print(
                f"{lesson.lesson_date}: "
                f"{lesson.teacher['short']} - "
                f"{lesson.lesson_name['short']}"
                f" - In auditorium: {lesson.auditorium}"
                if lesson.auditorium
                else ""
            )
            print(f"Card: {lesson.lesson_info}\n")



"""Enumerators"""

from enum import Enum, StrEnum


class RequestsEnum:
    """Contains information for Requests library"""

    class Methods(StrEnum):
        """Contains used HTTP Methods for requests"""

        GET = "GET"
        POST = "POST"

        @classmethod
        def choices(cls):
            """Returns list of available choices

            Returns:
                list[str]: list of available choices
            """
            return [cls.GET.value, cls.POST.value]

    class Codes(Enum):
        """Contains used HTTP response codes"""

        OK = 200
        SERVICE_UNAVAILABLE = 503

    @classmethod
    def code_ok(cls):
        """
        OK code

        Returns:
            int: 200, Codes.OK
        """
        return cls.Codes.OK.value

    @classmethod
    def method_get(cls):
        """Method GET

        Returns:
            str: GET
        """
        return cls.Methods.GET.value

    @classmethod
    def method_post(cls):
        """Method POST

        Returns:
            str: POST
        """
        return cls.Methods.POST.value



"""
Contains classes needed to get data
Like Faculty or Group, provides methods to get names, ids, etc.
"""

from urllib.parse import parse_qsl

from attrs import define
from bs4.element import Tag

from ontu_parser.classes.base import BaseClass


class BaseTag(BaseClass):
    """Base Tag Class for parsing BS4 tags from responses"""

    @classmethod
    def from_tag(cls, tag):
        """Checks tag and returns initialized object"""
        raise NotImplementedError("`from_tag` Not implemented")

    @staticmethod
    def _check_tag(tag: Tag):
        """Checks if tag is valid for usage"""
        raise NotImplementedError("`_check_tag` Not implemented")


@define
class Faculty(BaseTag):
    """Describes faculty from BS4 tag"""

    parent_id: int | None
    prefix: str
    faculty_tag: Tag

    @staticmethod
    def _check_tag(tag):
        attrs = getattr(tag, "attrs", None)
        span = getattr(tag, "span", None)
        required_properties = [attrs, span]
        if not all(required_properties):
            raise ValueError(f"Invalid tag: {tag}, has no attrs", tag)
        required = ["data-id"]
        for requirement in required:
            if requirement not in attrs:
                raise ValueError(
                    f"Invalid tag: {tag}, doesn't have attrs: {required}",
                    tag,
                    required,
                )
        span_string = getattr(span, "string", None)
        if span_string is None:
            raise ValueError(f"Invalid tag: {tag}, `span` has no string", tag)

    @classmethod
    def from_tag(
        cls,
        tag,
        prefix: str = "",
        parent_id: int | None = None,
    ):
        cls._check_tag(tag)
        return cls(
            faculty_tag=tag,
            prefix=prefix,
            parent_id=parent_id,
        )

    def get_faculty_picture(self):
        """Returns relative link to picture (if present)"""
        return self.faculty_tag.attrs.get("data-cover", None)

    def get_faculty_id(self):
        """Returns temporary id of faculty (for later use in search)"""
        return self.faculty_tag.attrs["data-id"]

    def get_faculty_name(self):
        """Returns name of the faculty"""
        return f"{self.prefix}{self.faculty_tag.span.string}"


@define
class Group(BaseTag):
    """Describes group from BS4 tag"""

    group_tag: Tag

    _icon_tag_filter = {"attrs": {"class": "icon"}}
    _text_tag_filter = {"attrs": {"class": "branding-bar"}}

    @staticmethod
    def _check_tag(tag):
        attrs = getattr(tag, "attrs", None)
        required = ["data-id"]
        for requirement in required:
            if requirement not in attrs:
                raise ValueError(
                    f"Invalid tag: {tag}, doesn't have attrs: {required}",
                    tag,
                    required,
                )

        # Children requiremenets

        icon = tag.find(**Group._icon_tag_filter)
        text = tag.find(**Group._text_tag_filter)
        required = [icon, text]
        if not all(required):
            raise ValueError(f"Invalid tag: {tag} doesn't have suitable children", tag)

    @classmethod
    def from_tag(cls, tag):
        cls._check_tag(tag)
        return cls(group_tag=tag)

    @property
    def text(self):
        """Returns text tag from group tag"""
        return self.group_tag.find(**self._text_tag_filter)

    @property
    def icon(self):
        """Returns icon tag from group tag"""
        return self.group_tag.find(**self._icon_tag_filter)

    def get_group_id(self):
        """Returns (temporary) id of this group"""
        return self.group_tag.attrs["data-id"]

    def get_group_name(self):
        """Retunrs a name of the group or None"""
        if not self.text:
            print(f"text tag not found in {self.group_tag}")
            return None
        return self.text.string

    def get_group_icon(self):
        """Returns name of the icon of the group or None"""
        if not self.icon:
            print(f"icon tag not found in {self.group_tag}")
            return None
        # Hardcoding this
        attrs = self.icon.attrs.copy()
        # Feels bad :(
        attrs.pop("icon")
        return attrs[0]


class BaseStudentsLesson(BaseTag):
    """
    Describes lesson from bs4 tag

    Note: Lesson is a concrete even with date and teacher
    Pair on the other hand just states at which time lesson will happen
    """

    lesson_tag: Tag

    lesson_date: str = ""
    lesson_info: str = ""
    auditorium: str | None = None

    def __init__(self, *args, **kwargs):
        self.teacher: dict = {}
        self.lesson_name: dict = {}
        super().__init__(*args, **kwargs)

    @staticmethod
    def _check_tag(tag: Tag):
        # Dear Gods, forgive me for not checking tags for lessons
        pass

    def parse_tag(self):
        """This method parses bs4 and stores data from it in object's fields"""
        raise NotImplementedError(
            "`parse_tag` was not implemented\n"
            "You are probably executing this from BaseLesson\n"
            "Please use one of derived classes"
        )


class StudentsRegularLesson(BaseStudentsLesson):
    """
    This class should be used to parse lesson from bs4 tag
    If you are getting schedule for current week
    """

    @classmethod
    def from_tag(cls, tag):
        obj = cls()
        obj.lesson_tag = tag
        obj.parse_tag()
        return obj

    def parse_tag(self):
        lesson_top = self.lesson_tag.parent

        predm_element = lesson_top.find(name="span", attrs={"class": "predm"})
        self.lesson_name = {
            "short": predm_element.text,
            "full": predm_element.attrs.get("title", "Not Set"),
        }

        prp_element = lesson_top.find(name="span", attrs={"class": "prp"})
        self.teacher = {
            "short": prp_element.text.replace("\xa0", " "),  # Why...
            "full": prp_element.attrs.get("title", "Not Set"),
        }

        # Card tag consists of two children
        # First states type of content
        # Other - content itself
        card_tag = lesson_top.find(name="div", attrs={"class": "card"})
        if card_tag:
            card_content = card_tag.find(name="div", attrs={"class": "card-content"})
            if card_content:
                self.lesson_info = card_content.text.replace("\t", "").strip()

        auditorium_tag = lesson_top.find(name="a", attrs={"class": "fg-blue"})
        if auditorium_tag:
            self.auditorium = auditorium_tag.text


class StudentsPair(BaseTag):
    """
    Describes pair from bs4 tag

    Note: Pair describes when certain Lesson will happen
    """

    pair_tag: Tag

    pair_no: int = None
    _subgroup_id: int = 0

    def __init__(self, *args, **kwargs):
        self.lessons: list[BaseStudentsLesson] = []
        super().__init__(*args, **kwargs)

    @staticmethod
    def _check_tag(tag: Tag):
        pass

    @classmethod
    def from_tag(cls, tag, subgroup_id=0):
        cls._check_tag(tag)
        obj = cls()
        obj._subgroup_id = subgroup_id

        obj.pair_tag = tag
        obj.set_pair_number()
        pair = obj.get_pair_tag_for_subgroup()
        lessons = cls.get_lessons(pair)
        obj.lessons = lessons
        return obj

    def set_pair_number(self):
        """This method gets pair number for better identification"""
        pair_no_tag = self.pair_tag.find(attrs={"class": "lesson"})
        self.pair_no = int(pair_no_tag.text)

    def get_pair_tag_for_subgroup(self):
        """
        This method returns tag for this pair accounting for subgroup
        Currently opening a page for subgroup (like KN-321[a]) opens
        a page for both subgroups (or a group), thus we have to get a correct cell
        """
        pair_no_tag = self.pair_tag.find(attrs={"class": "lesson"})
        skip = 1 + self._subgroup_id
        pair_tag = None
        for _ in range(skip):
            if not pair_tag:
                pair_tag = pair_no_tag.nextSibling
            else:
                pair_tag = pair_tag.nextSibling
        return pair_tag

    @staticmethod
    def get_lessons(pair: Tag):
        """Parses lessons for this pair"""
        # All time 'days' have <span>s with dates in them
        all_dates = pair.find_all(name="span", attrs={"class": "fg-blue"})
        # There is at least one tag with this class if
        # there are lessons
        lesson = pair.find(attrs={"class": "predm"})
        lessons = []
        if not any([len(all_dates), lesson]):
            return lessons
        if len(all_dates) > 0:
            # This means we are dealing with 'all time' records
            # Which have multiple lessons per pair
            for lesson in all_dates:
                lessons.append(StudentsRegularLesson.from_tag(lesson))
            return lessons
        # This means we are dealing with single week records
        lessons.append(StudentsRegularLesson.from_tag(lesson))
        return lessons


class BaseSchedule(BaseTag):
    """Describes schedule from BS4 tag"""

    schedule_tag: Tag

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._schedule_data: dict[str, list["StudentsPair | TeachersPair"]] = {}

    @property
    def week(self):
        """Gets data for this week"""
        self._get_week()
        return self._schedule_data

    def _get_week(self):
        raise NotImplementedError("`_get_week` was not implemented")


class StudentsSchedule(BaseSchedule):
    """Describes schedule from HTML table"""

    subgroup_id: int = 0
    _subgroup: str = ""

    _splitter_class: str = "bg-darkCyan"

    def __init__(self, *args, **kwargs):
        self.subgroups: list[str] = []
        super().__init__(*args, **kwargs)

    @staticmethod
    def _check_tag(tag):
        if tag.name != "table":
            raise ValueError(f"Invalid tag: {tag}. Should be table", tag)

    @classmethod
    def from_tag(cls, tag, subgroup=None):
        cls._check_tag(tag)
        obj = cls()
        obj.schedule_tag = tag

        obj._subgroup = subgroup
        obj._get_subgroup_id()

        return obj

    def _get_subgroup_id(self):
        if self._subgroup:
            if not self.subgroups:
                self._parse_subgroups()
            try:
                self.subgroup_id = self.subgroups.index(self._subgroup)
            except ValueError:
                print("Invalid subgroup! Please try making request again")

    def _parse_subgroups(self):
        """This method prepares subgroups for later use"""
        sub_groups_list = []
        table_head = self.schedule_tag.thead
        head_rows = table_head.find_all(name="tr")

        # Hardcoding positions! Yikes!
        # head_rows[0] - meta info (`Day`, `Pair` columns, Group name)
        # head_rows[1] - sub_groups (a/b etc)

        sub_groups_tag = head_rows[1]
        sub_groups_tags = sub_groups_tag.find_all(name="th")

        for sub_group in sub_groups_tags:
            sub_groups_list.append(sub_group.text.strip())

        self.subgroups = sub_groups_list

    def _prepare_day_tag(self, day_name_tag):
        """
        Parses day from 'day_name_tag'*
        Returns name of that day and a list of tags that represent pairs

        *day_name_tag is a tag that contains name of the tag
         It also has attr - class = day
        """
        pair_tags = []

        day_name: str = day_name_tag.text

        first_pair_tag = day_name_tag.parent
        # We also have to include this 'top tag', since it's first pair
        pair_tags.append(first_pair_tag)

        next_pair_tag = first_pair_tag.next_sibling
        # next_sibling gives next tag on the same level of hierarchy
        while True:
            if not next_pair_tag or isinstance(next_pair_tag, str):
                # We may not have next sibling
                # Or, as it happens RN - we may get '  ' as next tag :|
                break
            if self._splitter_class in next_pair_tag.attrs.get("class", []):
                # splitter has class `_splitter_class` (like bg-darkCyan)
                # if we hit splitter - day has ended
                break
            pair_tags.append(next_pair_tag)
            next_pair_tag = next_pair_tag.next_sibling
        if isinstance(pair_tags[-1], str):
            pair_tags.pop()
        return day_name, pair_tags

    def _prepare_tags(self, tags):
        """Parses bs4 tags to list of Pair objects"""
        prepared_tags: list[StudentsPair] = []
        for tag in tags:
            prepared_tags.append(
                StudentsPair.from_tag(tag, subgroup_id=self.subgroup_id)
            )
        return prepared_tags

    def _get_week(self):
        """Iteratively loops trough table to get data for all days"""
        table_body = self.schedule_tag.tbody
        days = table_body.find_all(attrs={"class": "day"})
        for day in days:
            day_name, tags = self._prepare_day_tag(day)
            prepared_days = self._prepare_tags(tags)
            self._schedule_data[day_name] = prepared_days
        return self._schedule_data


class TeachersLesson:
    """Class to describe lesson for teachers"""

    # pylint: disable=too-few-public-methods

    name: str
    groups: str

    def __init__(self, name: str, groups: list[str] | str) -> None:
        self.name = name
        if isinstance(groups, list):
            self.groups = ", ".join(groups)
        else:
            self.groups = groups

    def __str__(self) -> str:
        return f"Lesson: {self.name} with ({self.groups})"


class TeachersPair(BaseTag):
    """Describes pair from bs4 tag"""

    __pair_no_not_specified = "Не вказано"
    __pair_name_not_specified = "Назва не вказана"
    __groups_not_specified = "Групи не вказані"

    pair_tag: Tag
    pair_no: int
    lesson: TeachersLesson | None

    @staticmethod
    def _check_tag(tag: Tag):
        pass

    def parse_tag(self):
        """This method parses bs4 and stores data from it in object's fields"""
        pair_no_text = self.pair_tag.attrs.get(
            "data-title-caption", self.__pair_no_not_specified
        )
        if pair_no_text != self.__pair_no_not_specified:
            self.pair_no = int(pair_no_text.split()[0])
        else:
            self.pair_no = 0

        pair_name_tag = self.pair_tag.find(name="p", attrs={"class": "text-leader"})
        pair_name = pair_name_tag.text.strip() if pair_name_tag else None

        groups_tag = self.pair_tag.find(name="p", attrs={"class": "text-secondary"})
        # Consider splitting. e.g of content: КН-341[а], КН-342[а], КН-343[а], КН-343[б]
        groups = groups_tag.text.strip() if groups_tag else None

        lesson = None
        if pair_name or groups:
            lesson = TeachersLesson(
                name=pair_name or self.__pair_name_not_specified,
                groups=groups or self.__groups_not_specified,
            )

        self.lesson = lesson

    @classmethod
    def from_tag(cls, tag):
        cls._check_tag(tag)
        obj = cls()
        obj.pair_tag = tag
        obj.parse_tag()
        return obj


class TeacherSchedule(BaseSchedule):
    """Describes schedule from HTML grid"""

    @staticmethod
    def _check_tag(tag: Tag):
        if tag.name != "div":
            raise ValueError(f"Invalid tag: {tag}. Should be div", tag)
        if "grid" not in tag.attrs.get("class", []):
            raise ValueError(f"Invalid tag: {tag}. Should be grid", tag)

    @classmethod
    def from_tag(cls, tag):
        cls._check_tag(tag)
        obj = cls()
        obj.schedule_tag = tag
        return obj

    def _prepare_day_tag(self, day_card: "Tag"):
        day_name = day_card.find(name="div", attrs={"class": "card-header"})
        if not day_name:
            raise ValueError(f"Invalid tag: {day_card}. No card-header found", day_card)
        day_name = day_name.text.strip()
        pairs = []
        for pair in day_card.find_all(name="div", attrs={"data-role": "panel"}):
            pairs.append(pair)
        return day_name, pairs

    def _prepare_tags(self, tags: list["Tag"]):
        prepared_tags: list[TeachersPair] = []
        for tag in tags:
            prepared_tags.append(TeachersPair.from_tag(tag))
        return prepared_tags

    def _get_week(self):
        all_cards = self.schedule_tag.find_all(name="div", attrs={"class": "card"})
        for card in all_cards:
            day_name, tags = self._prepare_day_tag(card)
            prepared_days = self._prepare_tags(tags)
            self._schedule_data[day_name] = prepared_days
        return self._schedule_data


@define
class Department(BaseTag):
    """Describes department from BS4 tag"""

    department: Tag

    @staticmethod
    def _check_tag(tag):
        attrs = getattr(tag, "attrs", None)
        span = tag.find(name="span", attrs={"class": "branding-bar"})
        required_properties = [attrs, span]
        if not all(required_properties):
            raise ValueError(f"Invalid tag: {tag}, has no attrs", tag)
        required = ["href"]
        for requirement in required:
            if requirement not in attrs:
                raise ValueError(
                    f"Invalid tag: {tag}, doesn't have attrs: {required}",
                    tag,
                    required,
                )
        span_string = getattr(span, "string", None)
        if span_string is None:
            raise ValueError(f"Invalid tag: {tag}, `span` has no string", tag)

    @classmethod
    def from_tag(cls, tag):
        cls._check_tag(tag)
        obj = cls(department=tag)
        if not obj.department:
            raise ValueError("Invalid tag", tag)
        return obj

    def get_department_picture(self):
        """Returns class of the picture (if present)"""
        container = self.department.find(name="div", attrs={"class": "slide-front"})
        if not container:
            return None
        span = container.find(name="span")
        if not span:
            return None
        return span.attrs.get("class", None)

    def get_department_link(self):
        """Returns (semi?) permanent relative link to department"""
        return self.department.attrs["href"]

    def get_department_id(self) -> int:
        """Return id of the department"""
        key_dict = dict(parse_qsl(self.get_department_link()))
        return int(key_dict["dep"])

    def get_department_name(self) -> dict[str, str]:
        """Returns name of the faculty"""
        name = {"short": "", "full": ""}
        short_name_span = self.department.find(
            name="span", attrs={"class": "branding-bar"}
        )
        full_name_span = self.department.find(name="div", attrs={"class": "slide-back"})
        name["short"] = short_name_span.text.strip() if short_name_span else ""
        full_name = full_name_span.text.strip() if full_name_span else ""
        name["full"] = full_name
        if full_name:
            words = full_name.split()
            name["full"] = " ".join(
                [x.capitalize() if len(x) > 2 else x for x in words]
            )
        return name


@define
class Teacher(BaseTag):
    """Describes teacher from BS4 tag"""

    teacher: Tag

    @staticmethod
    def _check_tag(tag):
        attrs = getattr(tag, "attrs", None)
        span = tag.find(name="span", attrs={"class": "branding-bar"})
        required_properties = [attrs, span]
        if not all(required_properties):
            raise ValueError(f"Invalid tag: {tag}, has no attrs", tag)
        required = ["href"]
        for requirement in required:
            if requirement not in attrs:
                raise ValueError(
                    f"Invalid tag: {tag}, doesn't have attrs: {required}",
                    tag,
                    required,
                )
        span_string = getattr(span, "string", None)
        if span_string is None:
            raise ValueError(f"Invalid tag: {tag}, `span` has no string", tag)

    @classmethod
    def from_tag(cls, tag):
        cls._check_tag(tag)
        obj = cls(teacher=tag)
        if not obj.teacher:
            raise ValueError("Invalid tag", tag)
        return obj

    def get_teacher_picture(self):
        """Returns class of the picture (if present)"""
        container = self.teacher.find(name="div", attrs={"class": "slide-front"})
        if not container:
            return None
        span = container.find(name="span")
        if not span:
            return None
        return span.attrs.get("class", None)

    def get_teacher_link(self):
        """Returns (semi?) permanent relative link to department"""
        return self.teacher.attrs["href"]

    def get_teacher_id(self) -> int:
        """Return id of teacher"""
        key_dict = dict(parse_qsl(self.get_teacher_link()))
        return int(key_dict["teacher"])

    def get_teacher_name(self) -> dict[str, str]:
        """Returns name of the faculty"""
        name = {"short": "", "full": ""}
        short_name_span = self.teacher.find(
            name="span", attrs={"class": "branding-bar"}
        )
        full_name_span = self.teacher.find(name="div", attrs={"class": "slide-back"})
        name["short"] = short_name_span.text.strip() if short_name_span else ""
        full_name = full_name_span.text.strip() if full_name_span else ""
        name["full"] = full_name
        if full_name:
            words = full_name.split()
            name["full"] = " ".join(
                [x.capitalize() if len(x) > 2 else x for x in words]
            )
        return name



"""Module for parser class"""

from bs4 import BeautifulSoup
from requests import Response

from .base import BaseClass
from .dataclasses import (
    Department,
    Faculty,
    Group,
    StudentsSchedule,
    Teacher,
    TeacherSchedule,
)
from .enums import RequestsEnum
from .sender import Sender


class Parser(BaseClass):
    """Parser class to get information from Rozklad ONTU"""

    sender: Sender = None

    def __init__(self, *args, **kwargs):
        if isinstance(kwargs, dict) and "kwargs" in kwargs:
            # Trick to unwrap kwargs
            kwargs = kwargs.get("kwargs", {})

        self.sender = Sender(*args, **kwargs)

    def _get_page(self, response: Response):
        content = response.content
        if not content:
            raise ValueError(f"Response: {response} has no content!", response)
        decoded_content = content.decode("utf-8")
        return BeautifulSoup(decoded_content, "html.parser")

    def get_faculties(self) -> list[Faculty]:
        """Returns a list of faculties as Faculty objects"""
        faculties_response = self.sender.send_request(
            method=RequestsEnum.method_get()  # No data gives 'main' page with faculties
        )
        faculty_page = self._get_page(faculties_response)
        faculty_tags = faculty_page.find_all(
            attrs={"class": "fc"}
        )  # Faculties have class 'fc'
        faculty_entities = []
        for tag in faculty_tags:
            faculty_entities.append(Faculty.from_tag(tag))
        return faculty_entities

    def get_all_extramurals(self) -> list[Faculty]:
        """Returns a list of extramural faculties"""
        faculties = self.get_faculties()
        extramurals = []
        for faculty in faculties:
            extramural = self.get_extramural(faculty.get_faculty_id())
            if extramural:
                extramurals.append(extramural)

        return extramurals

    def get_extramural(self: "Parser", faculty_id: int) -> Faculty | None:
        """
        Returns extramural faculty by faculty id
        Returns None if no extramural faculty found
        """
        faculty_data = self.sender.send_request(
            method=RequestsEnum.method_post(),
            data={"facultyid": faculty_id},
        )
        faculty_page = self._get_page(faculty_data)
        faculty_tag = faculty_page.find(attrs={"class": "fc"})
        faculty_name_tag = faculty_page.find(attrs={"href": "?to_faculty=1"})
        if faculty_tag:
            return Faculty.from_tag(
                faculty_tag,
                prefix=(faculty_name_tag.text + " - ") if faculty_name_tag else "",
                parent_id=faculty_id,
            )
        return None

    def get_groups(
        self,
        faculty_id: int | str | None = None,
        faculty: Faculty | None = None,
    ) -> list[Group]:
        """Returns Group list of a faculty by faculty id"""
        if isinstance(faculty_id, str):
            faculty_id = int(faculty_id)
        if isinstance(faculty, Faculty):
            faculty_id = int(faculty.get_faculty_id())
            if faculty.parent_id:
                # Someone has decided that extramural groups can only be seen if you have seen this
                # specific parent first :shrug:
                self.get_groups(faculty_id=faculty.parent_id)

        if not any([faculty_id, faculty]):
            raise ValueError("Please specify one of the optional parameters")

        groups_response = self.sender.send_request(
            method=RequestsEnum.method_post(),
            data={"facultyid": faculty_id},
        )
        groups_page = self._get_page(groups_response)
        groups_tags = groups_page.find_all(attrs={"class": "grp"})
        group_entities: list[Group] = []
        for tag in groups_tags:
            group_entities.append(Group.from_tag(tag))
        return group_entities

    def get_schedule(
        self,
        group_id: int | None = None,
        teacher_id: int | None = None,
        all_time=False,
    ):
        """Returns schedule for group, or for teachers"""
        if group_id:
            return self._get_group_schedule(group_id, all_time=all_time).week
        if teacher_id:
            return self._get_teachers_schedule(teacher_id, all_time=all_time).week
        raise ValueError("No group or teacher id provided!")

    def _get_group_schedule(self, group_id, all_time=False) -> StudentsSchedule:
        """
        Returns a schedule for a group (by id)
        If all_time is False - returns schedule for current week
        Else - returns schedule for whole semester
        """
        request_data = {"groupid": group_id}
        if all_time:
            request_data["show_all"] = 1
        schedule_response = self.sender.send_request(
            method=RequestsEnum.method_post(),
            data=request_data,
        )
        schedule_page = self._get_page(schedule_response)

        breadcrumbs = schedule_page.find(attrs={"class": "breadcrumbs"})
        group_breadcrumbs = breadcrumbs.find_all(attrs={"class": "page-link"})

        table = schedule_page.find(attrs={"class": "table"})
        group_name = group_breadcrumbs[-1].text
        # I hate this, but at the same time - I love it
        # If it ever to become broken I'll implement this a bit thoughtfully :)
        subgroup_name = group_name.split("[")[1].replace("]", "")
        schedule = StudentsSchedule.from_tag(table, subgroup=subgroup_name)
        return schedule

    def _get_teachers_schedule(self, teacher_id, all_time=False) -> TeacherSchedule:
        """Returns a schedule for a teacher (by id)"""
        self._check_for_teachers()
        query = {"page": "teacher", "teacher": teacher_id}
        if all_time:
            query["page"] = "teacher_all"
            query["show"] = 1
        schedule_response = self.sender.send_request(
            method=RequestsEnum.method_get(),
            query=query,
        )
        schedule_page = self._get_page(schedule_response)

        grid = schedule_page.find(name="div", attrs={"class": "grid"})
        if not grid:
            raise ValueError("No grid found!")
        schedule = TeacherSchedule.from_tag(grid)
        return schedule

    def parse(self, all_time=False):
        """Parses information, requiring user input (CLI)"""
        schedule = None
        all_faculties = self.get_faculties()

        for faculty in all_faculties:
            print(faculty.get_faculty_name())

        faculty_name = input("Введите название факультета: ")
        faculty_id = None
        for faculty in all_faculties:
            if faculty.get_faculty_name() == faculty_name:
                faculty_id = faculty.get_faculty_id()
                break
        else:
            print("Несуществующее имя факльтета!")
            return schedule
        groups = self.get_groups(faculty_id)
        for group in groups:
            print(group.get_group_name())

        group_name = input("Введите название группы: ")
        group_id = None
        for group in groups:
            if group.get_group_name() == group_name:
                group_id = group.get_group_id()
                break
        else:
            print("Несуществующее имя группы!")
            return schedule

        schedule = self.get_schedule(group_id, all_time=all_time)
        return schedule

    def _check_for_teachers(self):
        """A check, that must be rune before executing teachers methods"""
        if not self.sender:
            raise ValueError("Sender is not set!")
        if not self.sender.for_teachers:
            raise ValueError("Sender is not set for teachers!")

    def get_departments(self) -> list["Department"]:
        """Returns a list of departments"""
        self._check_for_teachers()

        departments_response = self.sender.send_request(
            method=RequestsEnum.method_get()
        )
        departments_page = self._get_page(departments_response)
        titles = departments_page.find(attrs={"class": "tiles-grid"})
        if not titles:
            raise ValueError("No titles found!")
        departments_tags = titles.find_all(name="a", attrs={"data-role": "tile"})
        departments = []
        for tag in departments_tags:
            departments.append(Department.from_tag(tag))
        return departments

    def get_teachers_by_department(self, department_id: int) -> "list[Teacher]":
        """Returns a list of teachers by department id"""
        self._check_for_teachers()

        teachers_response = self.sender.send_request(
            method=RequestsEnum.method_get(),
            query={"page": "department", "dep": department_id},
        )
        teachers_page = self._get_page(teachers_response)
        teachers_tags = teachers_page.find_all(attrs={"class": "tiles-grid"})
        if not teachers_tags:
            raise ValueError("No teachers found!")
        teachers_tags = teachers_tags[0].find_all(name="a", attrs={"data-role": "tile"})
        teachers = []
        for tag in teachers_tags:
            teachers.append(Teacher.from_tag(tag))
        return teachers



"""
This is a module with classes capable of parsing and passing 2WAF JavaScript challenge.
"""

import re
import urllib.parse
from base64 import b64encode
from hashlib import sha256
from typing import TypedDict
import time

from bs4 import BeautifulSoup


class CookieValues(TypedDict):
    """
    Describes values of cookies
    """

    notbot: str
    php_session_id: str | None = None
    pow_result: str

    @staticmethod
    def to_dict(var: "CookieValues") -> dict[str, str]:
        """
        Converts CookieValues to a dictionary
        """
        value = {
            "notbot": var["notbot"],
            "pow-result": var["pow_result"],
        }
        if "php_session_id" in var:
            value["PHPSESSID"] = var["php_session_id"]
        return value


class JavaScriptParser:
    # pylint: disable=too-few-public-methods
    """
    Parses JavaScript from HTML to get necessary values
    """

    def __init__(self: "JavaScriptParser", html: str) -> None:
        self.html = html
        self.soup = BeautifulSoup(html, "html.parser")

        self._notbot_value: str = ""
        self._pow_result: str = ""

    def _extract_script_tags(self: "JavaScriptParser") -> list[str]:
        return [script.text for script in self.soup.find_all("script")]

    def _get_notbot_cookie(
        self: "JavaScriptParser",
        notbot_script: str,
    ) -> str:
        self._notbot_value = (
            notbot_script.split("setCookie('notbot','")[1].split("');")[0].strip()
        )
        return self._notbot_value

    def __make_combinations(
        self: "JavaScriptParser",
        characters: str,
        length: int,
    ) -> list[str]:
        if length == 0:
            return [""]
        return [
            char + combination
            for char in characters
            for combination in self.__make_combinations(characters, length - 1)
        ]

    def _get_pow_result(
        self: "JavaScriptParser",
        pow_script: str,
    ) -> str:
        pow_script = pow_script.replace("\n", "")
        expected_hash = re.search(
            r'const hash4find\s*=\s*["\'](.*?)["\']',
            pow_script,
            re.IGNORECASE,
        ).group(1)
        combination_characters = re.search(
            r'const chars\s*=\s*["\'](.*?)["\']',
            pow_script,
            re.IGNORECASE,
        ).group(1)
        combination_prefix = re.search(
            r'const prefix\s*=\s*["\'](.*?)["\']', pow_script, re.IGNORECASE
        ).group(1)
        combination_length = int(
            re.search(
                r"const suffixlen\s*=\s*(\d+)",
                pow_script,
                re.IGNORECASE,
            ).group(1)
        )

        for combination in self.__make_combinations(
            combination_characters,
            combination_length,
        ):
            hash_string = combination_prefix + combination
            hash_value = sha256(bytes(hash_string, "utf-8")).hexdigest()
            if hash_value == expected_hash:
                self._pow_result = b64encode(
                    urllib.parse.unquote_plus(urllib.parse.quote(hash_string)).encode()
                ).decode()
                return self._pow_result
        raise ValueError("Could not find pow_result")

    def parse(self: "JavaScriptParser") -> CookieValues:
        """
        This methods parses JavaScript from HTML and returns CookieValues object
        """
        start = time.time()
        script_tags = self._extract_script_tags()
        pow_result_script: str | None = None

        for script in script_tags:
            script = script.strip()

            if "document.onreadystatechange" in script:
                self._get_notbot_cookie(script)
                continue
            if '"use strict"' in script or "const hash4find" in script:
                pow_result_script = script
                continue
        if not pow_result_script:
            raise ValueError("Could not find pow_result script")

        self._get_pow_result(pow_result_script)
        end = time.time()

        # Stupid artificial delay posed by the website
        # 5.5 should be fine, but I'll add 0.5 just in case
        duration = end - start
        min_duration = 6
        if duration < min_duration:
            time.sleep(min_duration - duration)

        return CookieValues(
            notbot=self._notbot_value,
            pow_result=self._pow_result,
        )



"""Module for sending operations"""

import logging
import time
from datetime import datetime
from urllib.parse import urlencode

import requests

from ontu_parser.classes.js_parser import CookieValues, JavaScriptParser

from .base import BaseClass
from .enums import RequestsEnum

API_URL = "https://rozklad.ontu.edu.ua"
DEFAULT_HEADERS = {
    # Why not? I like to show off :)
    "User-Agent": "makisukurisu/rozklad-ontu-parser; (ontu_schedule_bot)",
}


class TTLValue(BaseClass):
    """Describes value with some time to live (like authorization token)"""

    _ttl: int = 3600  # Time To Live (in seconds)
    _value: object = None

    issued_at: datetime = datetime.min

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.issued_at = datetime.now()

    def is_valid(self):
        """
        Checks wether value is still valid
            True if TTLValue was issued less seconds before than Time To Live
        """
        seconds_passed = (datetime.now() - self.issued_at).seconds
        if seconds_passed < self._ttl:
            return True
        return False

    def set_value(self, value):
        """Sets value, and resets issued_at"""
        self.issued_at = datetime.now()
        self._value = value
        return self._value


class Cookies(TTLValue):
    """Describes cookies (Temporary values) for requests"""

    def __init__(self, sender, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sender: Sender = sender

    @property
    def value(self) -> dict[str, str]:
        """Returns value of a cookie (or gets one, if not present)"""
        if self._value and self.is_valid():
            return self._value

        logging.warning("Cookies are being updated")
        cookie = self.get_cookie()
        if not cookie:
            raise RuntimeError("Could not get cookies")
        return cookie

    def _try_get_cookie(self) -> dict[str, str]:
        with requests.session() as session:
            response = session.get(
                API_URL,
                headers=DEFAULT_HEADERS,
            )

            if response.status_code != RequestsEnum.Codes.SERVICE_UNAVAILABLE.value:
                logging.error("Expected to get a 503 response. Returning old value")
                return self._value

            cookies = JavaScriptParser(
                html=response.text,
            ).parse()

            session_response = session.get(
                API_URL,
                headers=DEFAULT_HEADERS,
                cookies=CookieValues.to_dict(cookies),
            )
            php_session_id = session_response.cookies.get("PHPSESSID")
            if (
                session_response.status_code
                == RequestsEnum.Codes.SERVICE_UNAVAILABLE.value
            ) or php_session_id is None:
                raise ValueError("Could not get cookies")

            cookies["PHPSESSID"] = php_session_id
            return self.set_value(cookies)

    def get_cookie(self):
        """
        Sends a request to a server. If we get response with status code 503,
        we know that we need to generate cookies.

        Cookies (at the moment) are generated by executing a
        JavaScript that's embedded in the 503 response.
        """
        # pylint: disable=logging-fstring-interpolation
        for attempt in range(1, 4):
            try:
                return self._try_get_cookie()
            except ValueError as exception:
                delay = attempt * attempt
                logging.error(
                    f"Could not get cookies, got exception: {exception}.",
                )
                logging.warning(
                    f"Retrying in {delay} seconds",
                )
                time.sleep(delay)
        raise RuntimeError("Could not get cookies after 3 attempts")


class Sender(BaseClass):
    """Describes sender with link and cookies to send requests"""

    default_link: str = f"{API_URL}/guest_n.php"
    teachers_link: str = f"{API_URL}/departments_all.php"
    cookies: Cookies = None
    for_teachers: bool = False

    @property
    def link(self):
        """Returns link to send requests to"""
        if self.for_teachers:
            return self.teachers_link
        return self.default_link

    def __init__(self, *args, **kwargs):
        for_teachers = kwargs.get("for_teachers", False)
        self.for_teachers = for_teachers
        self.cookies = Cookies(self)

    _responses: list[requests.Response] = []

    def send_request(
        self,
        method: str,
        data: (dict | None) = None,
        query: (dict | None) = None,
    ):
        """Sends request with method and some data, if needed"""
        session = requests.Session()
        if method not in RequestsEnum.Methods.choices():
            raise ValueError(
                f"arg. `method` should be one of: {RequestsEnum.Methods.choices()}",
                method,
            )

        link = self.link
        if query:
            link += "?" + urlencode(query)

        try:
            response: requests.Response = session.request(
                method=method,
                url=link,
                data=data,
                headers=DEFAULT_HEADERS,
                cookies=self.cookies.value,
            )
        except Exception as exception:
            raise ValueError(
                f"could not get response from {link}, got exception: {exception}"
            ) from exception
        if response.status_code != RequestsEnum.code_ok():
            raise ValueError(
                "server returned non OK response",
                response.status_code,
                response,
                response.content,
            )
        # Keep responses for a little while
        self._responses.append(response)
        self._responses = self._responses[-5:]

        return response
